@@ -5,6 +5,7 @@ import json
 import os
 import re
 import subprocess
+from itertools import product
 from typing import (
     Iterable,
     Optional,
@@ -234,6 +235,7 @@ DH_ADDONS_TO_REMOVE = frozenset(
         "doxygen",
         # Sequences that are embedded fully into debputy
         "bash-completion",
+        "shell-completions",
         "sodeps",
     ]
 )
@@ -473,6 +475,108 @@ def migrate_bash_completion(
                 )
 
 
+_SHELL_COMPLETIONS_RE = re.compile(r"^\s*\S+\s+\S+\s+\S")
+
+
+def migrate_shell_completions(
+    debian_dir: VirtualPath,
+    manifest: HighLevelManifest,
+    acceptable_migration_issues: AcceptableMigrationIssues,
+    feature_migration: FeatureMigration,
+    _migration_target: DebputyIntegrationMode,
+) -> None:
+    feature_migration.tagline = "dh_shell_completions files"
+    is_single_binary = sum(1 for _ in manifest.all_packages) == 1
+    mutable_manifest = assume_not_none(manifest.mutable_manifest)
+    installations = mutable_manifest.installations(create_if_absent=False)
+    # Note: The bash completion script used `bash-completion` whereas `dh_shell_completions` uses
+    # `...-completions` (note the trailing `s`). In `debputy`, we always use the singular notation
+    # because we use "one file, one completion (ruleset)".
+    completions = ["bash", "fish", "zsh"]
+
+    for completion, dctrl_bin in product(completions, manifest.all_packages):
+        dh_file = dhe_pkgfile(debian_dir, dctrl_bin, f"{completion}-completions")
+        if dh_file is None:
+            continue
+        is_completion_file = False
+        with dh_file.open() as fd:
+            for line in fd:
+                line = line.strip()
+                if not line or line[0] == "#":
+                    continue
+                if _SHELL_COMPLETIONS_RE.search(line):
+                    is_completion_file = True
+                    break
+        if is_completion_file:
+            dest_path = f"debian/{dctrl_bin.name}.{completion}-completion"
+            feature_migration.rename_on_success(dh_file.fs_path, dest_path)
+            continue
+
+        _, content = _dh_config_file(
+            debian_dir,
+            dctrl_bin,
+            f"{completion}-completions",
+            "dh_shell_completions",
+            acceptable_migration_issues,
+            feature_migration,
+            manifest,
+            remove_on_migration=True,
+        )
+
+        if content:
+            install_dest_sources: List[str] = []
+            install_as_rules: List[Tuple[str, str]] = []
+            for dhe_line in content:
+                if len(dhe_line.tokens) > 2:
+                    raise UnsupportedFeature(
+                        f"The dh_shell_completions file {dh_file.path} more than two words on"
+                        f' line {dhe_line.line_no} (line: "{dhe_line.original_line}").'
+                    )
+                source = dhe_line.tokens[0]
+                dest_basename = (
+                    dhe_line.tokens[1]
+                    if len(dhe_line.tokens) > 1
+                    else os.path.basename(source)
+                )
+                if source.startswith("debian/") and not has_glob_magic(source):
+                    if dctrl_bin.name != dest_basename:
+                        dest_path = f"debian/{dctrl_bin.name}.{dest_basename}.{completion}-completion"
+                    else:
+                        dest_path = f"debian/{dest_basename}.{completion}-completion"
+                    feature_migration.rename_on_success(source, dest_path)
+                elif len(dhe_line.tokens) == 1:
+                    install_dest_sources.append(source)
+                else:
+                    install_as_rules.append((source, dest_basename))
+
+            completion_dir_variable = (
+                "{{path:" + f"{completion.upper()}_COMPLETION_DIR" + "}}"
+            )
+
+            if install_dest_sources:
+                sources: Union[List[str], str] = (
+                    install_dest_sources
+                    if len(install_dest_sources) > 1
+                    else install_dest_sources[0]
+                )
+                installations.append(
+                    AbstractMutableYAMLInstallRule.install_dest(
+                        sources=sources,
+                        dest_dir=completion_dir_variable,
+                        into=dctrl_bin.name if not is_single_binary else None,
+                    )
+                )
+
+            for source, dest_basename in install_as_rules:
+                installations.append(
+                    AbstractMutableYAMLInstallRule.install_as(
+                        source=source,
+                        install_as=f"{completion_dir_variable}/{dest_basename}",
+                        into=dctrl_bin.name if not is_single_binary else None,
+                    )
+                )
+
+
 def migrate_dh_installsystemd_files(
     debian_dir: VirtualPath,
     manifest: HighLevelManifest,
