@@ -2,6 +2,7 @@ import dataclasses
 import os.path
 import re
 import textwrap
+from itertools import chain
 from typing import (
     Union,
     Sequence,
@@ -17,16 +18,16 @@ from debputy.analysis.analysis_util import flatten_ppfs
 from debputy.analysis.debian_dir import resolve_debhelper_config_files
 from debputy.dh.dh_assistant import extract_dh_compat_level
 from debputy.linting.lint_util import LintState
+from debputy.lsp.apt_cache import PackageLookup
 from debputy.lsp.debputy_ls import DebputyLanguageServer
 from debputy.lsp.diagnostics import DiagnosticData
 from debputy.lsp.lsp_debian_control_reference_data import (
     DctrlKnownField,
-    BINARY_FIELDS,
-    SOURCE_FIELDS,
     DctrlFileMetadata,
     package_name_to_section,
     all_package_relationship_fields,
     extract_first_value_and_position,
+    all_source_relationship_fields,
 )
 from debputy.lsp.lsp_features import (
     lint_diagnostics,
@@ -37,8 +38,9 @@ from debputy.lsp.lsp_features import (
     lsp_semantic_tokens_full,
     lsp_will_save_wait_until,
     lsp_format_document,
-    LanguageDispatch,
     lsp_text_doc_inlay_hints,
+    LanguageDispatchRule,
+    SecondaryLanguage,
 )
 from debputy.lsp.lsp_generic_deb822 import (
     deb822_completer,
@@ -99,7 +101,7 @@ from debputy.packager_provided_files import (
     detect_all_packager_provided_files,
 )
 from debputy.plugin.api.impl import plugin_metadata_for_debputys_own_plugin
-from debputy.util import detect_possible_typo
+from debputy.util import detect_possible_typo, PKGNAME_REGEX, _info
 
 try:
     from debputy.lsp.vendoring._deb822_repro.locatable import (
@@ -113,13 +115,16 @@ except ImportError:
     pass
 
 
-_LANGUAGE_IDS = [
-    LanguageDispatch.from_language_id("debian/control"),
-    # emacs's name
-    LanguageDispatch.from_language_id("debian-control"),
-    # vim's name
-    LanguageDispatch.from_language_id("debcontrol"),
-]
+_DISPATCH_RULE = LanguageDispatchRule.new_rule(
+    "debian/control",
+    "debian/control",
+    [
+        # emacs's name
+        SecondaryLanguage("debian-control"),
+        # vim's name
+        SecondaryLanguage("debcontrol"),
+    ],
+)
 
 
 @dataclasses.dataclass(slots=True, frozen=True)
@@ -291,7 +296,7 @@ _SUBSTVARS_DOC = _substvars_metadata(
             Note: This substvar only appears in `debhelper-compat (= 14)`, or
             with use of `debputy` (at an integration level, where `debputy`
             runs `dpkg-shlibdeps`), or when passing relevant options to
-            `dpkg-shlibdeps`  (often via `dh_shlibdeps`) such as `-dPre-Depends`.
+            `dpkg-shlibdeps` (often via `dh_shlibdeps`) such as `-dPre-Depends`.
     """
         ),
     ),
@@ -300,10 +305,10 @@ _SUBSTVARS_DOC = _substvars_metadata(
 _DCTRL_FILE_METADATA = DctrlFileMetadata()
 
 
-lsp_standard_handler(_LANGUAGE_IDS, TEXT_DOCUMENT_CODE_ACTION)
+lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)
 
 
-@lsp_hover(_LANGUAGE_IDS)
+@lsp_hover(_DISPATCH_RULE)
 def _debian_control_hover(
     ls: "DebputyLanguageServer",
     params: HoverParams,
@@ -311,10 +316,308 @@ def _debian_control_hover(
     return deb822_hover(ls, params, _DCTRL_FILE_METADATA, custom_handler=_custom_hover)
 
 
+def _custom_hover_description(
+    _ls: "DebputyLanguageServer",
+    _known_field: DctrlKnownField,
+    line: str,
+    _word_at_position: str,
+) -> Optional[Union[Hover, str]]:
+    if line[0].isspace():
+        return None
+    try:
+        col_idx = line.index(":")
+    except ValueError:
+        return None
+
+    content = line[col_idx + 1 :].strip()
+
+    # Synopsis
+    return textwrap.dedent(
+        f"""\
+        # Package synopsis
+
+        The synopsis functions as a phrase describing the package, not a
+        complete sentence, so sentential punctuation is inappropriate: it
+        does not need extra capital letters or a final period (full stop).
+        It should also omit any initial indefinite or definite article
+        - "a", "an", or "the". Thus for instance:
+
+        ```
+        Package: libeg0
+        Description: exemplification support library
+        ```
+
+        Technically this is a noun phrase minus articles, as opposed to a
+        verb phrase. A good heuristic is that it should be possible to
+        substitute the package name and synopsis into this formula:
+
+        ```
+        # Generic
+        The package provides {{a,an,the,some}} synopsis.
+
+        # The current package for comparison
+        The package provides {{a,an,the,some}} {content}.
+        ```
+
+        Other advice for writing synopsis:
+         * Avoid using the package name. Any software would display the
+           package name already and it generally does not help the user
+           understand what they are looking at.
+         * In many situations, the user will only see the package name
+           and its synopsis. The synopsis must be able to stand alone.
+
+        **Example renderings in various terminal UIs**:
+        ```
+        # apt search TERM
+        package/stable,now 1.0-1 all:
+           {content}
+
+        # apt-get search TERM
+        package - {content}
+        ```
+
+        ## Reference example
+
+        An reference example for comparison: The Sphinx package
+        (python3-sphinx/7.2.6-6) had the following synopsis:
+
+        ```
+        Description: documentation generator for Python projects
+        ```
+
+        In the test sentence, it would read as:
+
+        ```
+        The python3-sphinx package provides a documentation generator for Python projects.
+        ```
+
+        **Side-by-side comparison in the terminal UIs**:
+        ```
+        # apt search TERM
+        python3-sphinx/stable,now 7.2.6-6 all:
+           documentation generator for Python projects
+
+        package/stable,now 1.0-1 all:
+           {content}
+
+
+        # apt-get search TERM
+        package - {content}
+        python3-sphinx - documentation generator for Python projects
+        ```
+    """
+    )
+
+
+def _render_package_lookup(
+    package_lookup: PackageLookup,
+    known_field: DctrlKnownField,
+) -> str:
+    name = package_lookup.name
+    provider = package_lookup.package
+    if package_lookup.package is None and len(package_lookup.provided_by) == 1:
+        provider = package_lookup.provided_by[0]
+
+    if provider:
+        segments = [
+            f"# {name} ({provider.version}, {provider.architecture}) ",
+            "",
+        ]
+
+        if (
+            _is_bd_field(known_field)
+            and name.startswith("dh-sequence-")
+            and len(name) > 12
+        ):
+            sequence = name[12:]
+            segments.append(
+                f"This build-dependency will activate the `dh` sequence called `{sequence}`."
+            )
+            segments.append("")
+
+        elif (
+            known_field.name == "Build-Depends"
+            and name.startswith("debputy-plugin-")
+            and len(name) > 15
+        ):
+            plugin_name = name[15:]
+            segments.append(
+                f"This build-dependency will activate the `debputy` plugin called `{plugin_name}`."
+            )
+            segments.append("")
+
+        segments.extend(
+            [
+                f"Synopsis: {provider.synopsis}",
+                f"Multi-Arch: {provider.multi_arch}",
+                f"Section: {provider.section}",
+            ]
+        )
+        if provider.upstream_homepage is not None:
+            segments.append(f"Upstream homepage: {provider.upstream_homepage}")
+        segments.append("")
+        segments.append(
+            "Data is from the system's APT cache, which may not match the target distribution."
+        )
+        return "\n".join(segments)
+
+    segments = [
+        f"# {name} [virtual]",
+        "",
+        "The package {name} is a virtual package provided by one of:",
+    ]
+    segments.extend(f" * {p.name}" for p in package_lookup.provided_by)
+    segments.append("")
+    segments.append(
+        "Data is from the system's APT cache, which may not match the target distribution."
+    )
+    return "\n".join(segments)
+
+
+def _disclaimer(is_empty: bool) -> str:
+    if is_empty:
+        return textwrap.dedent(
+            """\
+        The system's APT cache is empty, so it was not possible to verify that the
+        package exist.
+"""
+        )
+    return textwrap.dedent(
+        """\
+        The package is not known by the APT cache on this system, so there may be typo
+        or the package may not be available in the version of your distribution.
+"""
+    )
+
+
+def _render_package_by_name(
+    name: str, known_field: DctrlKnownField, is_empty: bool
+) -> Optional[str]:
+    if _is_bd_field(known_field) and name.startswith("dh-sequence-") and len(name) > 12:
+        sequence = name[12:]
+        return (
+            textwrap.dedent(
+                f"""\
+        # {name}
+
+        This build-dependency will activate the `dh` sequence called `{sequence}`.
+
+        """
+            )
+            + _disclaimer(is_empty)
+        )
+    if (
+        known_field.name == "Build-Depends"
+        and name.startswith("debputy-plugin-")
+        and len(name) > 15
+    ):
+        plugin_name = name[15:]
+        return (
+            textwrap.dedent(
+                f"""\
+        # {name}
+
+        This build-dependency will activate the `debputy` plugin called `{plugin_name}`.
+
+        """
+            )
+            + _disclaimer(is_empty)
+        )
+    return (
+        textwrap.dedent(
+            f"""\
+        # {name}
+
+    """
+        )
+        + _disclaimer(is_empty)
+    )
+
+
+def _is_bd_field(known_field: DctrlKnownField) -> bool:
+    return known_field.name in (
+        "Build-Depends",
+        "Build-Depends-Arch",
+        "Build-Depends-Indep",
+    )
+
+
+def _custom_hover_relationship_field(
+    ls: "DebputyLanguageServer",
+    known_field: DctrlKnownField,
+    _line: str,
+    word_at_position: str,
+) -> Optional[Union[Hover, str]]:
+    apt_cache = ls.apt_cache
+    state = apt_cache.state
+    is_empty = False
+    _info(f"Rel field: {known_field.name} - {word_at_position} - {state}")
+    if "|" in word_at_position:
+        return textwrap.dedent(
+            f"""\
+            Sorry, no hover docs for OR relations at the moment.
+
+            The relation being matched: `{word_at_position}`
+
+            The code is missing logic to determine which side of the OR the lookup is happening.
+        """
+        )
+    match = next(iter(PKGNAME_REGEX.finditer(word_at_position)), None)
+    if match is None:
+        return
+    package = match.group()
+    if state == "empty-cache":
+        state = "loaded"
+        is_empty = True
+    if state == "loaded":
+        result = apt_cache.lookup(package)
+        if result is None:
+            return _render_package_by_name(
+                package,
+                known_field,
+                is_empty=is_empty,
+            )
+        return _render_package_lookup(result, known_field)
+
+    if state in (
+        "not-loaded",
+        "failed",
+        "tooling-not-available",
+    ):
+        details = apt_cache.load_error if apt_cache.load_error else "N/A"
+        return textwrap.dedent(
+            f"""\
+        Sorry, the APT cache data is not available due to an error or missing tool.
+
+        Details: {details}
+        """
+        )
+
+    if state == "empty-cache":
+        return f"Cannot lookup {package}: APT cache data was empty"
+
+    if state == "loading":
+        return f"Cannot lookup {package}: APT cache data is still being indexed. Please try again in a moment."
+    return None
+
+
+_CUSTOM_FIELD_HOVER = {
+    field: _custom_hover_relationship_field
+    for field in chain(
+        all_package_relationship_fields().values(),
+        all_source_relationship_fields().values(),
+    )
+    if field != "Provides"
+}
+
+_CUSTOM_FIELD_HOVER["Description"] = _custom_hover_description
+
+
 def _custom_hover(
+    ls: "DebputyLanguageServer",
     server_position: Position,
     _current_field: Optional[str],
-    _word_at_position: str,
+    word_at_position: str,
     known_field: Optional[DctrlKnownField],
     in_value: bool,
     _doc: "TextDocument",
@@ -361,96 +664,15 @@ def _custom_hover(
             md_fields = "\n" + substvar_md.render_metadata_fields()
         return f"# Substvar `{substvar}`\n\n{doc}{computed_doc}{md_fields}"
 
-    if known_field is None or known_field.name != "Description":
+    if known_field is None:
         return None
-    if line[0].isspace():
+    dispatch = _CUSTOM_FIELD_HOVER.get(known_field.name)
+    if dispatch is None:
         return None
-    try:
-        col_idx = line.index(":")
-    except ValueError:
-        return None
-
-    content = line[col_idx + 1 :].strip()
-
-    # Synopsis
-    return textwrap.dedent(
-        f"""\
-        # Package synopsis
-
-        The synopsis functions as a phrase describing the package, not a
-        complete sentence, so sentential punctuation is inappropriate: it
-        does not need extra capital letters or a final period (full stop).
-        It should also omit any initial indefinite or definite article
-        - "a", "an", or "the". Thus for instance:
-
-        ```
-        Package: libeg0
-        Description: exemplification support library
-        ```
-
-        Technically this is a noun phrase minus articles, as opposed to a
-        verb phrase. A good heuristic is that it should be possible to
-        substitute the package name and synopsis into this formula:
-
-        ```
-        # Generic
-        The package provides {{a,an,the,some}} synopsis.
-
-        # The current package for comparison
-        The package provides {{a,an,the,some}} {content}.
-        ```
-
-        Other advice for writing synopsis:
-         * Avoid using the package name. Any software would display the
-           package name already and it generally does not help the user
-           understand what they are looking at.
-         * In many situations, the user will only see the package name
-           and its synopsis. The synopsis must standalone.
-
-        **Example renderings in various terminal UIs**:
-        ```
-        # apt search TERM
-        package/stable,now 1.0-1 all:
-           {content}
-
-        # apt-get search TERM
-        package - {content}
-        ```
-
-        ## Reference example
-
-        An reference example for comparison: The Sphinx package
-        (python3-sphinx/7.2.6-6) had the following synopsis:
-
-        ```
-        Description: documentation generator for Python projects
-        ```
-
-        In the test sentence, it would read as:
-
-        ```
-        The python3-sphinx package provides a documentation generator for Python projects.
-        ```
-
-        **Side-by-side comparison in the terminal UIs**:
-        ```
-        # apt search TERM
-        python3-sphinx/stable,now 7.2.6-6 all:
-           documentation generator for Python projects
-
-        package/stable,now 1.0-1 all:
-           {content}
+    return dispatch(ls, known_field, line, word_at_position)
 
 
-        # apt-get search TERM
-        package - {content}
-        python3-sphinx - documentation generator for Python projects
-        ```
-    """
-    )
-
-
-@lsp_completer(_LANGUAGE_IDS)
+@lsp_completer(_DISPATCH_RULE)
 def _debian_control_completions(
     ls: "DebputyLanguageServer",
     params: CompletionParams,
@@ -458,7 +680,7 @@ def _debian_control_completions(
     return deb822_completer(ls, params, _DCTRL_FILE_METADATA)
 
 
-@lsp_folding_ranges(_LANGUAGE_IDS)
+@lsp_folding_ranges(_DISPATCH_RULE)
 def _debian_control_folding_ranges(
     ls: "DebputyLanguageServer",
     params: FoldingRangeParams,
@@ -466,7 +688,7 @@ def _debian_control_folding_ranges(
     return deb822_folding_ranges(ls, params, _DCTRL_FILE_METADATA)
 
 
-@lsp_text_doc_inlay_hints(_LANGUAGE_IDS)
+@lsp_text_doc_inlay_hints(_DISPATCH_RULE)
 def _doc_inlay_hint(
     ls: "DebputyLanguageServer",
     params: InlayHintParams,
@@ -486,7 +708,7 @@ def _doc_inlay_hint(
         stanza_range = stanza.range_in_parent()
         if stanza_no < 1:
             continue
-        pkg_kvpair = stanza.get_kvpair_element("Package", use_get=True)
+        pkg_kvpair = stanza.get_kvpair_element(("Package", 0), use_get=True)
         if pkg_kvpair is None:
             continue
 
@@ -542,12 +764,13 @@ def _source_package_checks(
     diagnostics: List[Diagnostic],
 ) -> None:
     vcs_fields = {}
+    source_fields = _DCTRL_FILE_METADATA["Source"].stanza_fields
     for kvpair in stanza.iter_parts_of_type(Deb822KeyValuePairElement):
         name = normalize_dctrl_field_name(kvpair.field_name.lower())
         if (
             not name.startswith("vcs-")
             or name == "vcs-browser"
-            or name not in SOURCE_FIELDS
+            or name not in source_fields
         ):
             continue
         vcs_fields[name] = kvpair
@@ -587,7 +810,7 @@ def _binary_package_checks(
 ) -> None:
     package_name = stanza.get("Package", "")
     source_section = source_stanza.get("Section")
-    section_kvpair = stanza.get_kvpair_element("Section", use_get=True)
+    section_kvpair = stanza.get_kvpair_element(("Section", 0), use_get=True)
     section: Optional[str] = None
     if section_kvpair is not None:
         section, section_range = extract_first_value_and_position(
@@ -976,7 +1199,7 @@ def _scan_for_syntax_errors_and_token_level_diagnostics(
     return first_error
 
 
-@lint_diagnostics(_LANGUAGE_IDS)
+@lint_diagnostics(_DISPATCH_RULE)
 def _lint_debian_control(
     lint_state: LintState,
 ) -> Optional[List[Diagnostic]]:
@@ -997,18 +1220,20 @@ def _lint_debian_control(
     source_paragraph = paragraphs[0] if paragraphs else None
     binary_stanzas_w_pos = []
 
+    source_stanza, binary_stanza = _DCTRL_FILE_METADATA.stanza_types()
+
     for paragraph_no, paragraph in enumerate(paragraphs, start=1):
         paragraph_pos = paragraph.position_in_file()
         if paragraph_pos.line_position >= first_error:
             break
         is_binary_paragraph = paragraph_no != 1
         if is_binary_paragraph:
-            known_fields = BINARY_FIELDS
-            other_known_fields = SOURCE_FIELDS
+            known_fields = binary_stanza.stanza_fields
+            other_known_fields = source_stanza.stanza_fields
             binary_stanzas_w_pos.append((paragraph, paragraph_pos))
         else:
-            known_fields = SOURCE_FIELDS
-            other_known_fields = BINARY_FIELDS
+            known_fields = source_stanza.stanza_fields
+            other_known_fields = binary_stanza.stanza_fields
         _diagnostics_for_paragraph(
             deb822_file,
             paragraph,
@@ -1036,7 +1261,7 @@ def _package_range_of_stanza(
     binary_stanzas: List[Tuple[Deb822ParagraphElement, TEPosition]],
 ) -> Iterable[Tuple[str, Optional[str], Range]]:
     for stanza, stanza_position in binary_stanzas:
-        kvpair = stanza.get_kvpair_element("Package", use_get=True)
+        kvpair = stanza.get_kvpair_element(("Package", 0), use_get=True)
         if kvpair is None:
             continue
         representation_field_range = kvpair.range_in_parent().relative_to(
@@ -1235,7 +1460,7 @@ def _detect_misspelled_packaging_files(
             )
 
 
-@lsp_will_save_wait_until(_LANGUAGE_IDS)
+@lsp_will_save_wait_until(_DISPATCH_RULE)
 def _debian_control_on_save_formatting(
     ls: "DebputyLanguageServer",
     params: WillSaveTextDocumentParams,
@@ -1251,7 +1476,7 @@ def _reformat_debian_control(
     return deb822_format_file(lint_state, _DCTRL_FILE_METADATA)
 
 
-@lsp_format_document(_LANGUAGE_IDS)
+@lsp_format_document(_DISPATCH_RULE)
 def _debian_control_on_save_formatting(
     ls: "DebputyLanguageServer",
     params: DocumentFormattingParams,
@@ -1261,7 +1486,7 @@ def _debian_control_on_save_formatting(
     return deb822_format_file(lint_state, _DCTRL_FILE_METADATA)
 
 
-@lsp_semantic_tokens_full(_LANGUAGE_IDS)
+@lsp_semantic_tokens_full(_DISPATCH_RULE)
 def _debian_control_semantic_tokens_full(
     ls: "DebputyLanguageServer",
     request: SemanticTokensParams,
