@@ -20,13 +20,19 @@ from debputy.lsp.lsp_features import (
     lsp_standard_handler,
     lsp_hover,
     lsp_completer,
-    LanguageDispatch,
+    SecondaryLanguage,
+    LanguageDispatchRule,
 )
 from debputy.lsp.lsp_generic_yaml import (
     resolve_hover_text,
     as_hover_doc,
     is_before,
     word_range_at_position,
+    YAML_COMPLETION_HINT_VALUE,
+    YAML_COMPLETION_HINT_KEY,
+    insert_complete_marker_snippet,
+    yaml_key_range,
+    yaml_flag_unknown_key,
 )
 from debputy.lsp.quickfixes import propose_correct_text_quick_fix
 from debputy.lsp.text_util import (
@@ -47,17 +53,16 @@ from debputy.lsprotocol.types import (
     DiagnosticRelatedInformation,
     Location,
 )
-from debputy.manifest_parser.base_types import DebputyDispatchableType
 from debputy.manifest_parser.declarative_parser import (
     AttributeDescription,
     ParserGenerator,
     DeclarativeNonMappingInputParser,
 )
 from debputy.manifest_parser.declarative_parser import DeclarativeMappingInputParser
+from debputy.manifest_parser.tagging_types import DebputyDispatchableType
 from debputy.manifest_parser.util import AttributePath
 from debputy.plugin.api.impl import plugin_metadata_for_debputys_own_plugin
 from debputy.plugin.api.impl_types import (
-    OPARSER_MANIFEST_ROOT,
     DeclarativeInputParser,
     DispatchingParserBase,
     DebputyPluginMetadata,
@@ -65,6 +70,7 @@ from debputy.plugin.api.impl_types import (
     InPackageContextParser,
     DeclarativeValuelessKeywordInputParser,
 )
+from debputy.plugin.api.parser_tables import OPARSER_MANIFEST_ROOT
 from debputy.plugin.api.spec import DebputyIntegrationMode
 from debputy.plugin.debputy.private_api import Capability, load_libcap
 from debputy.util import _info, detect_possible_typo
@@ -85,21 +91,22 @@ except ImportError:
     pass
 
 
-_LANGUAGE_IDS = [
-    LanguageDispatch.from_language_id("debian/debputy.manifest"),
-    LanguageDispatch.from_language_id("debputy.manifest"),
-    # LSP's official language ID for YAML files
-    LanguageDispatch.from_language_id(
-        "yaml", filename_selector="debian/debputy.manifest"
-    ),
-]
+_DISPATCH_RULE = LanguageDispatchRule.new_rule(
+    "debian/debputy.manifest",
+    "debian/debputy.manifest",
+    [
+        SecondaryLanguage("debputy.manifest"),
+        # LSP's official language ID for YAML files
+        SecondaryLanguage("yaml", filename_based_lookup=True),
+    ],
+)
 
 
-lsp_standard_handler(_LANGUAGE_IDS, TEXT_DOCUMENT_CODE_ACTION)
-lsp_standard_handler(_LANGUAGE_IDS, TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)
+lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)
+lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)
 
 
-@lint_diagnostics(_LANGUAGE_IDS)
+@lint_diagnostics(_DISPATCH_RULE)
 def _lint_debian_debputy_manifest(
     lint_state: LintState,
 ) -> Optional[List[Diagnostic]]:
@@ -176,7 +183,7 @@ def _integration_mode_allows_key(
         return
     if debputy_integration_mode in expected_debputy_integration_modes:
         return
-    key_range = _key_range(key, line, col, lines, position_codec)
+    key_range = yaml_key_range(key, line, col, lines, position_codec)
     yield Diagnostic(
         key_range,
         f'Feature "{key}" not supported in integration mode {debputy_integration_mode}',
@@ -185,64 +192,6 @@ def _integration_mode_allows_key(
     )
 
 
-def _key_range(
-    key: str,
-    line: int,
-    col: int,
-    lines: List[str],
-    position_codec: LintCapablePositionCodec,
-) -> Range:
-    key_len = len(key) if key else 1
-    return position_codec.range_to_client_units(
-        lines,
-        Range(
-            Position(
-                line,
-                col,
-            ),
-            Position(
-                line,
-                col + key_len,
-            ),
-        ),
-    )
-
-
-def _unknown_key(
-    key: Optional[str],
-    expected_keys: Iterable[str],
-    line: int,
-    col: int,
-    lines: List[str],
-    position_codec: LintCapablePositionCodec,
-    *,
-    message_format: str = 'Unknown or unsupported key "{key}".',
-) -> Tuple["Diagnostic", Optional[str]]:
-    key_range = _key_range(key, line, col, lines, position_codec)
-
-    candidates = detect_possible_typo(key, expected_keys) if key is not None else ()
-    extra = ""
-    corrected_key = None
-    if candidates:
-        extra = f' It looks like a typo of "{candidates[0]}".'
-        # TODO: We should be able to tell that `install-doc` and `install-docs` are the same.
-        #  That would enable this to work in more cases.
-        corrected_key = candidates[0] if len(candidates) == 1 else None
-
-    if key is None:
-        message_format = "Missing key"
-    diagnostic = Diagnostic(
-        key_range,
-        message_format.format(key=key) + extra,
-        DiagnosticSeverity.Error,
-        source="debputy",
-        data=DiagnosticData(
-            quickfixes=[propose_correct_text_quick_fix(n) for n in candidates]
-        ),
-    )
-    return diagnostic, corrected_key
-
-
 def _conflicting_key(
     uri: str,
     key_a: str,
@@ -412,7 +361,7 @@ def _lint_declarative_mapping_input_parser(
         attr = parser.manifest_attributes.get(key)
         line, col = lc.key(key)
         if attr is None:
-            diag, corrected_key = _unknown_key(
+            diag, corrected_key = yaml_flag_unknown_key(
                 key,
                 parser.manifest_attributes,
                 line,
@@ -488,15 +437,17 @@ def _lint_content(
             line, col = lc.key(key)
             orig_key = key
             if not is_known:
-                diag, corrected_key = _unknown_key(
+                diag, corrected_key = yaml_flag_unknown_key(
                     key,
                     parser.registered_keywords(),
                     line,
                     col,
                     lint_state.lines,
                     lint_state.position_codec,
+                    allow_unknown_keys=parser.diagnostics_allow_unknown_keys,
                 )
-                yield diag
+                if diag is not None:
+                    yield diag
                 if corrected_key is not None:
                     key = corrected_key
                     is_known = True
@@ -536,7 +487,7 @@ def _lint_content(
         for k, v in content.items():
             if "{{" not in k and known_packages is not None and k not in known_packages:
                 line, col = lc.key(k)
-                diag, _ = _unknown_key(
+                diag, _ = yaml_flag_unknown_key(
                     k,
                     known_packages,
                     line,
@@ -545,6 +496,7 @@ def _lint_content(
                     lint_state.position_codec,
                     message_format='Unknown package "{key}".',
                 )
+                assert diag is not None
                 yield diag
             yield from _lint_content(
                 lint_state,
@@ -605,10 +557,6 @@ def _trace_cursor(
     return None
 
 
-_COMPLETION_HINT_KEY = "___COMPLETE:"
-_COMPLETION_HINT_VALUE = "___COMPLETE"
-
-
 def resolve_keyword(
     current_parser: Union[DeclarativeInputParser[Any], DispatchingParserBase],
     current_plugin: DebputyPluginMetadata,
@@ -639,7 +587,7 @@ def resolve_keyword(
         return None
 
     if is_completion_attempt and current_segment.endswith(
-        (_COMPLETION_HINT_KEY, _COMPLETION_HINT_VALUE)
+        (YAML_COMPLETION_HINT_KEY, YAML_COMPLETION_HINT_VALUE)
     ):
         return current_parser, current_plugin, segment_idx
 
@@ -698,51 +646,23 @@ def resolve_keyword(
 DEBPUTY_PLUGIN_METADATA = plugin_metadata_for_debputys_own_plugin()
 
 
-def _escape(v: str) -> str:
-    return '"' + v.replace("\n", "\\n") + '"'
+def _maybe_quote(v: str) -> str:
+    if v and v[0].isdigit():
+        try:
+            float(v)
+            return f"'{v}'"
+        except ValueError:
+            pass
+    return v
 
 
-def _insert_snippet(lines: List[str], server_position: Position) -> bool:
-    _info(f"Complete at {server_position}")
-    line_no = server_position.line
-    line = lines[line_no]
-    pos_rhs = line[server_position.character :]
-    if pos_rhs and not pos_rhs.isspace():
-        _info(f"No insertion: {_escape(line[server_position.character:])}")
-        return False
-    lhs_ws = line[: server_position.character]
-    lhs = lhs_ws.strip()
-    if lhs.endswith(":"):
-        _info("Insertion of value (key seen)")
-        new_line = line[: server_position.character] + _COMPLETION_HINT_VALUE + "\n"
-    elif lhs.startswith("-"):
-        _info("Insertion of key or value (list item)")
-        # Respect the provided indentation
-        snippet = _COMPLETION_HINT_KEY if ":" not in lhs else _COMPLETION_HINT_VALUE
-        new_line = line[: server_position.character] + snippet + "\n"
-    elif not lhs or (lhs_ws and not lhs_ws[0].isspace()):
-        _info(f"Insertion of key or value: {_escape(line[server_position.character:])}")
-        # Respect the provided indentation
-        snippet = _COMPLETION_HINT_KEY if ":" not in lhs else _COMPLETION_HINT_VALUE
-        new_line = line[: server_position.character] + snippet + "\n"
-    elif lhs.isalpha() and ":" not in lhs:
-        _info(f"Expanding value to a key: {_escape(line[server_position.character:])}")
-        # Respect the provided indentation
-        new_line = line[: server_position.character] + _COMPLETION_HINT_KEY + "\n"
-    else:
-        c = (
-            line[server_position.character]
-            if server_position.character < len(line)
-            else "(OOB)"
-        )
-        _info(f"Not touching line: {_escape(line)} -- {_escape(c)}")
-        return False
-    _info(f'Evaluating complete on synthetic line: "{new_line}"')
-    lines[line_no] = new_line
-    return True
+def _complete_value(v: Any) -> str:
+    if isinstance(v, str):
+        return _maybe_quote(v)
+    return str(v)
 
 
-@lsp_completer(_LANGUAGE_IDS)
+@lsp_completer(_DISPATCH_RULE)
 def debputy_manifest_completer(
     ls: "DebputyLanguageServer",
     params: CompletionParams,
@@ -752,10 +672,9 @@ def debputy_manifest_completer(
     server_position = doc.position_codec.position_from_client_units(
         lines, params.position
     )
-    attribute_root_path = AttributePath.root_path()
     orig_line = lines[server_position.line].rstrip()
     has_colon = ":" in orig_line
-    added_key = _insert_snippet(lines, server_position)
+    added_key = insert_complete_marker_snippet(lines, server_position)
     attempts = 1 if added_key else 2
     content = None
 
@@ -783,7 +702,9 @@ def debputy_manifest_completer(
 
             if attempts > 0:
                 # Try to make it a key and see if that fixes the problem
-                new_line = lines[server_position.line].rstrip() + _COMPLETION_HINT_KEY
+                new_line = (
+                    lines[server_position.line].rstrip() + YAML_COMPLETION_HINT_KEY
+                )
                 lines[server_position.line] = new_line
         except YAMLError:
             break
@@ -791,6 +712,7 @@ def debputy_manifest_completer(
         context = lines[server_position.line].replace("\n", "\\n")
         _info(f"Completion failed: parse error: Line in question: {context}")
         return None
+    attribute_root_path = AttributePath.root_path(content)
     m = _trace_cursor(content, attribute_root_path, server_position)
 
     if m is None:
@@ -820,7 +742,9 @@ def debputy_manifest_completer(
         if isinstance(parser, DispatchingParserBase):
             if matched_key:
                 items = [
-                    CompletionItem(k if has_colon else f"{k}:")
+                    CompletionItem(
+                        _maybe_quote(k) if has_colon else f"{_maybe_quote(k)}:"
+                    )
                     for k in parser.registered_keywords()
                     if k not in parent
                     and not isinstance(
@@ -830,7 +754,7 @@ def debputy_manifest_completer(
                 ]
             else:
                 items = [
-                    CompletionItem(k)
+                    CompletionItem(_maybe_quote(k))
                     for k in parser.registered_keywords()
                     if k not in parent
                     and isinstance(
@@ -842,7 +766,9 @@ def debputy_manifest_completer(
             binary_packages = ls.lint_state(doc).binary_packages
             if binary_packages is not None:
                 items = [
-                    CompletionItem(p if has_colon else f"{p}:")
+                    CompletionItem(
+                        _maybe_quote(p) if has_colon else f"{_maybe_quote(p)}:"
+                    )
                     for p in binary_packages
                     if p not in parent
                 ]
@@ -858,7 +784,9 @@ def debputy_manifest_completer(
                         locked.add(attr_name)
                         break
                 items = [
-                    CompletionItem(k if has_colon else f"{k}:")
+                    CompletionItem(
+                        _maybe_quote(k) if has_colon else f"{_maybe_quote(k)}:"
+                    )
                     for k in parser.manifest_attributes
                     if k not in locked
                 ]
@@ -913,11 +841,11 @@ def _completion_from_attr(
         _info(f"Already filled: {matched} is one of {valid_values}")
         return None
     if valid_values:
-        return [CompletionItem(x) for x in valid_values]
+        return [CompletionItem(_complete_value(x)) for x in valid_values]
     return None
 
 
-@lsp_hover(_LANGUAGE_IDS)
+@lsp_hover(_DISPATCH_RULE)
 def debputy_manifest_hover(
     ls: "DebputyLanguageServer",
     params: HoverParams,
@@ -925,13 +853,13 @@ def debputy_manifest_hover(
     doc = ls.workspace.get_text_document(params.text_document.uri)
     lines = doc.lines
     position_codec = doc.position_codec
-    attribute_root_path = AttributePath.root_path()
     server_position = position_codec.position_from_client_units(lines, params.position)
 
     try:
         content = MANIFEST_YAML.load("".join(lines))
     except YAMLError:
         return None
+    attribute_root_path = AttributePath.root_path(content)
     m = _trace_cursor(content, attribute_root_path, server_position)
     if m is None:
         _info("No match")
