@@ -9,35 +9,12 @@ from typing import (
     Dict,
 )
 
-from debputy.lsprotocol.types import (
-    DiagnosticSeverity,
-    Range,
-    Diagnostic,
-    Position,
-    CompletionItem,
-    CompletionList,
-    CompletionParams,
-    DiagnosticRelatedInformation,
-    Location,
-    HoverParams,
-    Hover,
-    TEXT_DOCUMENT_CODE_ACTION,
-    SemanticTokens,
-    SemanticTokensParams,
-    FoldingRangeParams,
-    FoldingRange,
-    WillSaveTextDocumentParams,
-    TextEdit,
-    DocumentFormattingParams,
-)
-
 from debputy.linting.lint_util import LintState
 from debputy.lsp.debputy_ls import DebputyLanguageServer
 from debputy.lsp.diagnostics import DiagnosticData
 from debputy.lsp.lsp_debian_control_reference_data import (
     Deb822KnownField,
     DTestsCtrlFileMetadata,
-    _DTESTSCTRL_FIELDS,
 )
 from debputy.lsp.lsp_features import (
     lint_diagnostics,
@@ -48,7 +25,8 @@ from debputy.lsp.lsp_features import (
     lsp_semantic_tokens_full,
     lsp_will_save_wait_until,
     lsp_format_document,
-    LanguageDispatch,
+    SecondaryLanguage,
+    LanguageDispatchRule,
 )
 from debputy.lsp.lsp_generic_deb822 import (
     deb822_completer,
@@ -75,6 +53,27 @@ from debputy.lsp.vendoring._deb822_repro.parsing import (
     Deb822KeyValuePairElement,
     LIST_SPACE_SEPARATED_INTERPRETATION,
 )
+from debputy.lsprotocol.types import (
+    DiagnosticSeverity,
+    Range,
+    Diagnostic,
+    Position,
+    CompletionItem,
+    CompletionList,
+    CompletionParams,
+    DiagnosticRelatedInformation,
+    Location,
+    HoverParams,
+    Hover,
+    TEXT_DOCUMENT_CODE_ACTION,
+    SemanticTokens,
+    SemanticTokensParams,
+    FoldingRangeParams,
+    FoldingRange,
+    WillSaveTextDocumentParams,
+    TextEdit,
+    DocumentFormattingParams,
+)
 from debputy.util import detect_possible_typo
 
 try:
@@ -91,20 +90,23 @@ except ImportError:
 
 
 _CONTAINS_SPACE_OR_COLON = re.compile(r"[\s:]")
-_LANGUAGE_IDS = [
-    LanguageDispatch.from_language_id("debian/tests/control"),
-    # emacs's name - expected in elpa-dpkg-dev-el (>> 37.11)
-    LanguageDispatch.from_language_id("debian-autopkgtest-control-mode"),
-    # Likely to be vim's name if it had support
-    LanguageDispatch.from_language_id("debtestscontrol"),
-]
+_DISPATCH_RULE = LanguageDispatchRule.new_rule(
+    "debian/tests/control",
+    "debian/tests/control",
+    [
+        # emacs's name - from elpa-dpkg-dev-el (>> 37.11)
+        SecondaryLanguage("debian-autopkgtest-control-mode"),
+        # Likely to be vim's name if it had support
+        SecondaryLanguage("debtestscontrol"),
+    ],
+)
 
 _DTESTS_CTRL_FILE_METADATA = DTestsCtrlFileMetadata()
 
-lsp_standard_handler(_LANGUAGE_IDS, TEXT_DOCUMENT_CODE_ACTION)
+lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)
 
 
-@lsp_hover(_LANGUAGE_IDS)
+@lsp_hover(_DISPATCH_RULE)
 def debian_tests_control_hover(
     ls: "DebputyLanguageServer",
     params: HoverParams,
@@ -112,7 +114,7 @@ def debian_tests_control_hover(
     return deb822_hover(ls, params, _DTESTS_CTRL_FILE_METADATA)
 
 
-@lsp_completer(_LANGUAGE_IDS)
+@lsp_completer(_DISPATCH_RULE)
 def debian_tests_control_completions(
     ls: "DebputyLanguageServer",
     params: CompletionParams,
@@ -120,7 +122,7 @@ def debian_tests_control_completions(
     return deb822_completer(ls, params, _DTESTS_CTRL_FILE_METADATA)
 
 
-@lsp_folding_ranges(_LANGUAGE_IDS)
+@lsp_folding_ranges(_DISPATCH_RULE)
 def debian_tests_control_folding_ranges(
     ls: "DebputyLanguageServer",
     params: FoldingRangeParams,
@@ -430,7 +432,7 @@ def _scan_for_syntax_errors_and_token_level_diagnostics(
     return first_error
 
 
-@lint_diagnostics(_LANGUAGE_IDS)
+@lint_diagnostics(_DISPATCH_RULE)
 def _lint_debian_tests_control(
     lint_state: LintState,
 ) -> Optional[List[Diagnostic]]:
@@ -453,7 +455,10 @@ def _lint_debian_tests_control(
         paragraph_pos = paragraph.position_in_file()
         if paragraph_pos.line_position >= first_error:
             break
-        known_fields = _DTESTSCTRL_FIELDS
+        known_fields = _DTESTS_CTRL_FILE_METADATA.classify_stanza(
+            paragraph,
+            paragraph_no,
+        ).stanza_fields
         _diagnostics_for_paragraph(
             deb822_file,
             paragraph,
@@ -466,7 +471,7 @@ def _lint_debian_tests_control(
     return diagnostics
 
 
-@lsp_will_save_wait_until(_LANGUAGE_IDS)
+@lsp_will_save_wait_until(_DISPATCH_RULE)
 def _debian_tests_control_on_save_formatting(
     ls: "DebputyLanguageServer",
     params: WillSaveTextDocumentParams,
@@ -482,7 +487,7 @@ def _reformat_debian_tests_control(
     return deb822_format_file(lint_state, _DTESTS_CTRL_FILE_METADATA)
 
 
-@lsp_format_document(_LANGUAGE_IDS)
+@lsp_format_document(_DISPATCH_RULE)
 def _debian_tests_control_on_save_formatting(
     ls: "DebputyLanguageServer",
     params: DocumentFormattingParams,
@@ -492,7 +497,7 @@ def _debian_tests_control_on_save_formatting(
     return deb822_format_file(lint_state, _DTESTS_CTRL_FILE_METADATA)
 
 
-@lsp_semantic_tokens_full(_LANGUAGE_IDS)
+@lsp_semantic_tokens_full(_DISPATCH_RULE)
 def _debian_tests_control_semantic_tokens_full(
     ls: "DebputyLanguageServer",
     request: SemanticTokensParams,
