@@ -1,4 +1,3 @@
-import dataclasses
 import re
 from itertools import chain
 from typing import (
@@ -155,7 +154,7 @@ def _allow_stanza_continuation(
 
 def _at_cursor(
     deb822_file: Deb822FileElement,
-    file_metadata: Deb822FileMetadata[S],
+    file_metadata: Deb822FileMetadata[S, F],
     doc: "TextDocument",
     lines: List[str],
     client_position: Position,
@@ -241,7 +240,7 @@ def _at_cursor(
 def deb822_completer(
     ls: "DebputyLanguageServer",
     params: CompletionParams,
-    file_metadata: Deb822FileMetadata[Any],
+    file_metadata: Deb822FileMetadata[Any, Any],
 ) -> Optional[Union[CompletionList, Sequence[CompletionItem]]]:
     doc = ls.workspace.get_text_document(params.text_document.uri)
     lines = doc.lines
@@ -306,7 +305,7 @@ def deb822_completer(
 def deb822_hover(
     ls: "DebputyLanguageServer",
     params: HoverParams,
-    file_metadata: Deb822FileMetadata[S],
+    file_metadata: Deb822FileMetadata[S, F],
     *,
     custom_handler: Optional[
         Callable[
@@ -378,15 +377,15 @@ def deb822_hover(
             keyword = known_field.known_values.get(word_at_position)
             if keyword is None:
                 return None
-            hover_text = keyword.hover_text
+            hover_text = keyword.long_description_translated(ls)
             if hover_text is not None:
                 header = "`{VALUE}` (Field: {FIELD_NAME})".format(
                     VALUE=keyword.value,
                     FIELD_NAME=known_field.name,
                 )
-                hover_text = f"# {header})\n\n{hover_text}"
+                hover_text = f"# {header}\n\n{hover_text}"
         else:
-            hover_text = known_field.hover_text
+            hover_text = known_field.long_description_translated(ls)
             if hover_text is None:
                 hover_text = (
                     f"No documentation is available for the field {current_field}."
@@ -432,7 +431,7 @@ def deb822_folding_ranges(
     ls: "DebputyLanguageServer",
     params: FoldingRangeParams,
     # Unused for now: might be relevant for supporting folding for some fields
-    _file_metadata: Deb822FileMetadata[Any],
+    _file_metadata: Deb822FileMetadata[Any, Any],
 ) -> Optional[Sequence[FoldingRange]]:
     doc = ls.workspace.get_text_document(params.text_document.uri)
     comment_start = -1
@@ -475,7 +474,7 @@ class Deb822SemanticTokensState(SemanticTokensState):
         doc: "TextDocument",
         lines: List[str],
         tokens: List[int],
-        file_metadata: Deb822FileMetadata[Any],
+        file_metadata: Deb822FileMetadata[Any, Any],
         keyword_token_code: int,
         known_value_token_code: int,
         comment_token_code: int,
@@ -573,7 +572,7 @@ def _deb822_paragraph_semantic_tokens_full(
 
 def deb822_format_file(
     lint_state: LintState,
-    file_metadata: Deb822FileMetadata[Any],
+    file_metadata: Deb822FileMetadata[Any, Any],
 ) -> Optional[Sequence[TextEdit]]:
     effective_preference = lint_state.effective_preference
     if effective_preference is None:
@@ -600,7 +599,7 @@ def deb822_format_file(
 def deb822_semantic_tokens_full(
     ls: "DebputyLanguageServer",
     request: SemanticTokensParams,
-    file_metadata: Deb822FileMetadata[Any],
+    file_metadata: Deb822FileMetadata[Any, Any],
 ) -> Optional[SemanticTokens]:
     doc = ls.workspace.get_text_document(request.text_document.uri)
     position_codec = doc.position_codec
