@@ -1,11 +1,26 @@
 import dataclasses
 import textwrap
-from typing import Optional, Union, Mapping, Sequence, Callable, Iterable, Literal
+from typing import (
+    Optional,
+    Union,
+    Mapping,
+    Sequence,
+    Callable,
+    Iterable,
+    Literal,
+    Any,
+    Self,
+    TYPE_CHECKING,
+)
 
+from debputy.lsp.ref_models.deb822_reference_parse_models import UsageHint
 from debputy.lsp.vendoring._deb822_repro import Deb822ParagraphElement
 
+if TYPE_CHECKING:
+    from debputy.linting.lint_util import LintState
+    from debputy.lsp.debputy_ls import DebputyLanguageServer
 
-UsageHint = Literal["rare",]
+LSP_DATA_DOMAIN = "debputy-lsp-data"
 
 
 def format_comp_item_synopsis_doc(
@@ -21,8 +36,9 @@ def format_comp_item_synopsis_doc(
 @dataclasses.dataclass(slots=True, frozen=True)
 class Keyword:
     value: str
-    synopsis_doc: Optional[str] = None
-    hover_text: Optional[str] = None
+    synopsis: Optional[str] = None
+    long_description: Optional[str] = None
+    translation_context: str = ""
     is_obsolete: bool = False
     replaced_by: Optional[str] = None
     is_exclusive: bool = False
@@ -49,6 +65,25 @@ class Keyword:
             or self.can_complete_keyword_in_stanza(stanza_parts)
         )
 
+    def replace(self, **changes: Any) -> "Self":
+        return dataclasses.replace(self, **changes)
+
+    def synopsis_translated(
+        self, translation_provider: Union["DebputyLanguageServer", "LintState"]
+    ) -> str:
+        return translation_provider.translation(LSP_DATA_DOMAIN).pgettext(
+            self.translation_context,
+            self.synopsis,
+        )
+
+    def long_description_translated(
+        self, translation_provider: Union["DebputyLanguageServer", "LintState"]
+    ) -> str:
+        return translation_provider.translation(LSP_DATA_DOMAIN).pgettext(
+            self.translation_context,
+            self.long_description,
+        )
+
 
 def allowed_values(*values: Union[str, Keyword]) -> Mapping[str, Keyword]:
     as_keywords = [k if isinstance(k, Keyword) else Keyword(k) for k in values]
@@ -63,7 +98,7 @@ def allowed_values(*values: Union[str, Keyword]) -> Mapping[str, Keyword]:
 ALL_PUBLIC_NAMED_STYLES = allowed_values(
     Keyword(
         "black",
-        hover_text=textwrap.dedent(
+        long_description=textwrap.dedent(
             """\
             Uncompromising file formatting of Debian packaging files
 
