@@ -140,7 +140,9 @@ def _find_definition(
             continue
         # If the stem is also the extension and a known one at that, then
         # we do not consider it a typo match (to avoid false positives).
-        if not had_arch and stem in _KNOWN_NON_TYPO_EXTENSIONS:
+        #
+        # We also ignore "foo.1" since manpages are kind of common.
+        if not had_arch and (stem in _KNOWN_NON_TYPO_EXTENSIONS or stem.isdigit()):
             continue
         matches = detect_possible_typo(stem, stems)
         if matches is not None and len(matches) == 1:
@@ -413,7 +415,7 @@ def detect_all_packager_provided_files(
         period2stems = {}
 
     for entry in debian_dir.iterdir:
-        if entry.is_dir:
+        if entry.is_dir or entry.name.startswith("."):
             continue
         if entry.path in ignore_paths or entry.name in _KNOWN_NON_PPFS:
             continue
