@@ -1,6 +1,5 @@
 import dataclasses
 import os.path
-import textwrap
 from typing import (
     Optional,
     Callable,
@@ -24,22 +23,21 @@ from typing import (
     Set,
     Iterator,
     Container,
+    Protocol,
 )
 from weakref import ref
 
-from debputy import DEBPUTY_DOC_ROOT_DIR
 from debputy.exceptions import (
     DebputyFSIsROError,
     PluginAPIViolationError,
     PluginConflictError,
     UnhandledOrUnexpectedErrorFromPluginError,
+    PluginBaseError,
+    PluginInitializationError,
 )
 from debputy.filesystem_scan import as_path_def
-from debputy.installations import InstallRule
-from debputy.maintscript_snippet import DpkgMaintscriptHelperCommand
-from debputy.manifest_conditions import ManifestCondition
-from debputy.manifest_parser.base_types import DebputyParsedContent, TypeMapping
 from debputy.manifest_parser.exceptions import ManifestParseException
+from debputy.manifest_parser.tagging_types import DebputyParsedContent, TypeMapping
 from debputy.manifest_parser.util import AttributePath, check_integration_mode
 from debputy.packages import BinaryPackage
 from debputy.plugin.api import (
@@ -62,15 +60,16 @@ from debputy.plugin.api.spec import (
     TypeMappingDocumentation,
     DebputyIntegrationMode,
 )
+from debputy.plugin.plugin_state import (
+    run_in_context_of_plugin,
+)
 from debputy.substitution import VariableContext
-from debputy.transformation_rules import TransformationRule
 from debputy.util import _normalize_path, package_cross_check_precheck
 
 if TYPE_CHECKING:
     from debputy.plugin.api.spec import (
         ServiceDetector,
         ServiceIntegrator,
-        PackageTypeSelector,
     )
     from debputy.manifest_parser.parser_data import ParserContextData
     from debputy.highlevel_manifest import (
@@ -78,10 +77,10 @@ if TYPE_CHECKING:
         PackageTransformationDefinition,
         BinaryPackageData,
     )
-
-
-_PACKAGE_TYPE_DEB_ONLY = frozenset(["deb"])
-_ALL_PACKAGE_TYPES = frozenset(["deb", "udeb"])
+    from debputy.plugin.debputy.to_be_api_types import (
+        BuildSystemRule,
+        BuildRuleParsedFormat,
+    )
 
 
 TD = TypeVar("TD", bound="Union[DebputyParsedContent, List[DebputyParsedContent]]")
@@ -89,26 +88,12 @@ PF = TypeVar("PF")
 SF = TypeVar("SF")
 TP = TypeVar("TP")
 TTP = Type[TP]
+BSR = TypeVar("BSR", bound="BuildSystemRule")
 
 DIPKWHandler = Callable[[str, AttributePath, "ParserContextData"], TP]
 DIPHandler = Callable[[str, PF, AttributePath, "ParserContextData"], TP]
 
 
-def resolve_package_type_selectors(
-    package_type: "PackageTypeSelector",
-) -> FrozenSet[str]:
-    if package_type is _ALL_PACKAGE_TYPES or package_type is _PACKAGE_TYPE_DEB_ONLY:
-        return cast("FrozenSet[str]", package_type)
-    if isinstance(package_type, str):
-        return (
-            _PACKAGE_TYPE_DEB_ONLY
-            if package_type == "deb"
-            else frozenset([package_type])
-        )
-    else:
-        return frozenset(package_type)
-
-
 @dataclasses.dataclass(slots=True)
 class DebputyPluginMetadata:
     plugin_name: str
@@ -143,7 +128,17 @@ class DebputyPluginMetadata:
     def load_plugin(self) -> None:
         plugin_loader = self.plugin_loader
         assert plugin_loader is not None
-        self.plugin_initializer = plugin_loader()
+        try:
+            self.plugin_initializer = run_in_context_of_plugin(
+                self.plugin_name,
+                plugin_loader,
+            )
+        except PluginBaseError:
+            raise
+        except Exception as e:
+            raise PluginInitializationError(
+                f"Initialization of {self.plugin_name} failed due to its initializer raising an exception"
+            ) from e
         assert self.plugin_initializer is not None
 
 
@@ -270,12 +265,10 @@ class MetadataOrMaintscriptDetector:
                 " this stage (file system layout is committed and the attempted changes"
                 " would be lost)."
             ) from e
-        except (ChildProcessError, RuntimeError, AttributeError) as e:
-            nv = f"{self.plugin_metadata.plugin_name}"
-            raise UnhandledOrUnexpectedErrorFromPluginError(
-                f"The plugin {nv} threw an unhandled or unexpected exception from its metadata"
-                f" detector with id {self.detector_id}."
-            ) from e
+        except UnhandledOrUnexpectedErrorFromPluginError as e:
+            e.add_note(
+                f"The exception was raised by the detector with the ID: {self.detector_id}"
+            )
 
 
 class DeclarativeInputParser(Generic[TD]):
@@ -383,6 +376,10 @@ class DispatchingParserBase(Generic[TP]):
         self.manifest_attribute_path_template = manifest_attribute_path_template
         self._parsers: Dict[str, PluginProvidedParser[Any, TP]] = {}
 
+    @property
+    def diagnostics_allow_unknown_keys(self) -> bool:
+        return False
+
     def is_known_keyword(self, keyword: str) -> bool:
         return keyword in self._parsers
 
@@ -494,6 +491,7 @@ class DispatchingObjectParser(
         expected_debputy_integration_mode: Optional[
             Container[DebputyIntegrationMode]
         ] = None,
+        diagnostics_allow_unknown_keys: bool = False,
     ) -> None:
         super().__init__(manifest_attribute_path_template)
         self._attribute_documentation: List[ParserAttributeDocumentation] = []
@@ -501,6 +499,11 @@ class DispatchingObjectParser(
             parser_documentation = reference_documentation()
         self._parser_documentation = parser_documentation
         self._expected_debputy_integration_mode = expected_debputy_integration_mode
+        self._diagnostics_allow_unknown_keys = diagnostics_allow_unknown_keys
+
+    @property
+    def diagnostics_allow_unknown_keys(self) -> bool:
+        return self._diagnostics_allow_unknown_keys
 
     @property
     def expected_debputy_integration_mode(
@@ -587,11 +590,11 @@ class DispatchingObjectParser(
             )
         if not isinstance(orig_value, dict):
             raise ManifestParseException(
-                f"The attribute {attribute_path.path} must be a non-empty mapping.{doc_ref}"
+                f"The attribute {attribute_path.path_container_lc} must be a non-empty mapping.{doc_ref}"
             )
         if not orig_value:
             raise ManifestParseException(
-                f"The attribute {attribute_path.path} must be a non-empty mapping.{doc_ref}"
+                f"The attribute {attribute_path.path_container_lc} must be a non-empty mapping.{doc_ref}"
             )
         result = {}
         unknown_keys = orig_value.keys() - self._parsers.keys()
@@ -605,7 +608,7 @@ class DispatchingObjectParser(
                 )
             remaining_valid_attribute_names = ", ".join(remaining_valid_attributes)
             raise ManifestParseException(
-                f'The attribute "{first_key}" is not applicable at {attribute_path.path}(with the current set'
+                f'The attribute "{first_key}" is not applicable at {attribute_path.path} (with the current set'
                 " of plugins). Possible attributes available (and not already used) are:"
                 f" {remaining_valid_attribute_names}.{doc_ref}"
             )
@@ -615,7 +618,10 @@ class DispatchingObjectParser(
             if value is None:
                 if isinstance(provided_parser.parser, DispatchingObjectParser):
                     provided_parser.handler(
-                        key, {}, attribute_path[key], parser_context
+                        key,
+                        {},
+                        attribute_path[key],
+                        parser_context,
                     )
                 continue
             value_path = attribute_path[key]
@@ -675,7 +681,7 @@ class InPackageContextParser(
             )
         if not isinstance(orig_value, dict) or not orig_value:
             raise ManifestParseException(
-                f"The attribute {attribute_path.path} must be a non-empty mapping.{doc_ref}"
+                f"The attribute {attribute_path.path_container_lc} must be a non-empty mapping.{doc_ref}"
             )
         delegate = self.delegate
         result = {}
@@ -774,64 +780,6 @@ class DeclarativeValuelessKeywordInputParser(DeclarativeInputParser[None]):
         )
 
 
-SUPPORTED_DISPATCHABLE_TABLE_PARSERS = {
-    InstallRule: "installations",
-    TransformationRule: "packages.{{PACKAGE}}.transformations",
-    DpkgMaintscriptHelperCommand: "packages.{{PACKAGE}}.conffile-management",
-    ManifestCondition: "*.when",
-}
-
-OPARSER_MANIFEST_ROOT = "<ROOT>"
-OPARSER_PACKAGES_ROOT = "packages"
-OPARSER_PACKAGES = "packages.{{PACKAGE}}"
-OPARSER_MANIFEST_DEFINITIONS = "definitions"
-
-SUPPORTED_DISPATCHABLE_OBJECT_PARSERS = {
-    OPARSER_MANIFEST_ROOT: reference_documentation(
-        reference_documentation_url=f"{DEBPUTY_DOC_ROOT_DIR}/MANIFEST-FORMAT.md",
-    ),
-    OPARSER_MANIFEST_DEFINITIONS: reference_documentation(
-        title="Packager provided definitions",
-        description="Reusable packager provided definitions such as manifest variables.",
-        reference_documentation_url=f"{DEBPUTY_DOC_ROOT_DIR}/MANIFEST-FORMAT.md#packager-provided-definitions",
-    ),
-    OPARSER_PACKAGES: reference_documentation(
-        title="Binary package rules",
-        description=textwrap.dedent(
-            """\
-            Inside the manifest, the `packages` mapping can be used to define requests for the binary packages
-            you want `debputy` to produce.  Each key inside `packages` must be the name of a binary package
-            defined in `debian/control`.  The value is a dictionary defining which features that `debputy`
-            should apply to that binary package.  An example could be:
-
-                packages:
-                    foo:
-                        transformations:
-                            - create-symlink:
-                                  path: usr/share/foo/my-first-symlink
-                                  target: /usr/share/bar/symlink-target
-                            - create-symlink:
-                                  path: usr/lib/{{DEB_HOST_MULTIARCH}}/my-second-symlink
-                                  target: /usr/lib/{{DEB_HOST_MULTIARCH}}/baz/symlink-target
-                    bar:
-                        transformations:
-                        - create-directories:
-                           - some/empty/directory.d
-                           - another/empty/integration-point.d
-                        - create-directories:
-                             path: a/third-empty/directory.d
-                             owner: www-data
-                             group: www-data
-
-            In this case, `debputy` will create some symlinks inside the `foo` package and some directories for
-            the `bar` package.  The following subsections define the keys you can use under each binary package.
-        """
-        ),
-        reference_documentation_url=f"{DEBPUTY_DOC_ROOT_DIR}/MANIFEST-FORMAT.md#binary-package-rules",
-    ),
-}
-
-
 @dataclasses.dataclass(slots=True)
 class PluginProvidedManifestVariable:
     plugin_metadata: DebputyPluginMetadata
@@ -1214,6 +1162,11 @@ class PluginProvidedKnownPackagingFile:
     plugin_metadata: DebputyPluginMetadata
 
 
+class BuildSystemAutoDetector(Protocol):
+
+    def __call__(self, source_root: VirtualPath, *args: Any, **kwargs: Any) -> bool: ...
+
+
 @dataclasses.dataclass(slots=True, frozen=True)
 class PluginProvidedTypeMapping:
     mapped_type: TypeMapping[Any, Any]
@@ -1221,6 +1174,19 @@ class PluginProvidedTypeMapping:
     plugin_metadata: DebputyPluginMetadata
 
 
+@dataclasses.dataclass(slots=True, frozen=True)
+class PluginProvidedBuildSystemAutoDetection(Generic[BSR]):
+    manifest_keyword: str
+    build_system_rule_type: Type[BSR]
+    detector: BuildSystemAutoDetector
+    constructor: Callable[
+        ["BuildRuleParsedFormat", AttributePath, "HighLevelManifest"],
+        BSR,
+    ]
+    auto_detection_shadow_build_systems: FrozenSet[str]
+    plugin_metadata: DebputyPluginMetadata
+
+
 class PackageDataTable:
     def __init__(self, package_data_table: Mapping[str, "BinaryPackageData"]) -> None:
         self._package_data_table = package_data_table
