@@ -19,6 +19,11 @@ from lint_tests.lint_tutil import (
 from debputy.lsprotocol.types import Diagnostic, DiagnosticSeverity
 from tutil import build_time_only
 
+ALL_REQUIRED_FILES = [
+    "debian/changelog",
+    "debian/copyright",
+]
+
 
 class DctrlLintWrapper(LintWrapper):
 
@@ -435,7 +440,7 @@ def test_dctrl_lint_sv(line_linter: LintWrapper) -> None:
     assert diagnostics and len(diagnostics) == 1
     issue = diagnostics[0]
 
-    msg = f'Not a valid version. Current version is "{CURRENT_STANDARDS_VERSION}"'
+    msg = f'Not a valid standards version. Current version is "{CURRENT_STANDARDS_VERSION}"'
     assert issue.message == msg
     assert f"{issue.range}" == "3:19-3:23"
     assert issue.severity == DiagnosticSeverity.Warning
@@ -709,6 +714,74 @@ def test_dctrl_lint_synopsis_too_short(line_linter: LintWrapper) -> None:
     assert issue.severity == DiagnosticSeverity.Warning
 
 
+def test_dctrl_lint_description_line_too_long(line_linter: LintWrapper) -> None:
+    lines = textwrap.dedent(
+        f"""\
+    Source: foo
+    Section: devel
+    Priority: optional
+    Standards-Version: {CURRENT_STANDARDS_VERSION}
+    Maintainer: Jane Developer <jane@example.com>
+    Build-Depends: debhelper-compat (= 13)
+
+    Package: foo
+    Architecture: all
+    Depends: bar, baz
+    Description: some short synopsis
+     A very interesting description
+     with a acceptable synopsis
+     .
+     This description has a line that is waaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaay too long
+     .
+     Just so be clear, this is for a test.
+    """
+    ).splitlines(keepends=True)
+
+    diagnostics = line_linter(lines)
+    print(diagnostics)
+    assert diagnostics and len(diagnostics) == 1
+    issue = diagnostics[0]
+
+    msg = "Package description line is too long; please line wrap it."
+    assert issue.message == msg
+    assert f"{issue.range}" == "14:79-14:96"
+    assert issue.severity == DiagnosticSeverity.Warning
+
+
+def test_dctrl_lint_description_template(line_linter: LintWrapper) -> None:
+    lines = textwrap.dedent(
+        f"""\
+    Source: foo
+    Section: devel
+    Priority: optional
+    Standards-Version: {CURRENT_STANDARDS_VERSION}
+    Maintainer: Jane Developer <jane@example.com>
+    Build-Depends: debhelper-compat (= 13)
+
+    Package: foo
+    Architecture: all
+    Depends: bar, baz
+    Description: some short synopsis
+     A very interesting description
+     with a acceptable synopsis
+     .
+     A wild <insert long description, indented with spaces> appeared
+     .
+     Just so be clear, this is for a test.
+    """
+    ).splitlines(keepends=True)
+
+    diagnostics = line_linter(lines)
+    print(diagnostics)
+    assert diagnostics and len(diagnostics) == 1
+    issue = diagnostics[0]
+
+    msg = "Unfilled or left-over template from dh_make"
+    assert issue.message == msg
+    assert f"{issue.range}" == "14:8-14:55"
+    assert issue.severity == DiagnosticSeverity.Error
+
+
 @build_time_only
 def test_dctrl_lint_ambiguous_pkgfile(line_linter: LintWrapper) -> None:
     lines = textwrap.dedent(
@@ -737,6 +810,8 @@ def test_dctrl_lint_ambiguous_pkgfile(line_linter: LintWrapper) -> None:
         [
             virtual_path_def(".", fs_path="."),
             "./debian/bar.service",
+            *ALL_REQUIRED_FILES,
+            "debian/rules",
         ]
     )
 
@@ -788,6 +863,8 @@ def test_dctrl_lint_ambiguous_pkgfile_no_name_segment(line_linter: LintWrapper)
         [
             virtual_path_def(".", fs_path="."),
             "./debian/bar.alternatives",
+            *ALL_REQUIRED_FILES,
+            "debian/rules",
         ]
     )
 
@@ -840,6 +917,8 @@ def test_dctrl_lint_stem_typo_pkgfile(line_linter: LintWrapper) -> None:
         [
             virtual_path_def(".", fs_path="."),
             "./debian/foo.intsall",
+            *ALL_REQUIRED_FILES,
+            "debian/rules",
         ]
     )
 
@@ -902,6 +981,7 @@ def test_dctrl_lint_stem_inactive_pkgfile_fp(line_linter: LintWrapper) -> None:
             """
                 ),
             ),
+            *ALL_REQUIRED_FILES,
         ]
     )
 
@@ -941,6 +1021,7 @@ def test_dctrl_lint_stem_typo_pkgfile_ignored_exts_or_files(
     line_linter.source_root = build_virtual_file_system(
         [
             virtual_path_def(".", fs_path="."),
+            "debian/.control.swp",
             "debian/salsa-ci.yml",
             "debian/gbp.conf",
             "debian/foo.conf",
@@ -948,6 +1029,8 @@ def test_dctrl_lint_stem_typo_pkgfile_ignored_exts_or_files(
             "debian/foo.yml",
             # One wrong one to ensure the test works.
             "debian/foo.intsall",
+            *ALL_REQUIRED_FILES,
+            "debian/rules",
         ]
     )
 
@@ -1209,3 +1292,72 @@ def test_dctrl_lint_dep_field_restricted_or_relations(
     assert problem_text == "|"
     assert f"{issue.range}" == "13:1-13:2"
     assert issue.severity == DiagnosticSeverity.Error
+
+
+def test_dctrl_duplicate_key(line_linter: LintWrapper) -> None:
+    lines = textwrap.dedent(
+        f"""\
+        Source: jquery-tablesorter
+        Section: javascript
+        Priority: optional
+        Maintainer: Debian Javascript Maintainers <pkg-javascript-devel@lists.alioth.de\
+        bian.org>
+        Uploaders: Paul Gevers <elbrus@debian.org>
+        Build-Depends:
+         debhelper-compat (=13),
+         grunt,
+         libjs-qunit,
+         node-grunt-contrib-clean,
+         node-grunt-contrib-copy,
+         node-grunt-contrib-uglify,
+         node-grunt-contrib-concat,
+        Standards-Version: {CURRENT_STANDARDS_VERSION}
+        Homepage: https://github.com/Mottie/tablesorter
+        Vcs-Git: https://salsa.debian.org/js-team/jquery-tablesorter.git
+        Vcs-Browser: https://salsa.debian.org/js-team/jquery-tablesorter
+        Rules-Requires-Root: no
+
+        Package: libjs-jquery-tablesorter
+        Architecture: all
+        Multi-Arch: foreign
+        Depends:
+         ${{misc:Depends}},
+         libjs-jquery,
+         libjs-jquery-metadata,
+        Recommends: javascript-common
+        Multi-Arch: foreign
+        Description: jQuery flexible client-side table sorting plugin
+         Tablesorter is a jQuery plugin for turning a standard HTML table with THEAD
+         and TBODY tags into a sortable table without page refreshes. Tablesorter can
+         successfully parse and sort many types of data including linked data in a
+         cell. It has many useful features including:
+         .
+           * Multi-column alphanumeric sorting and filtering.
+           * Multi-tbody sorting
+           * Supports Bootstrap v2-4.
+           * Parsers for sorting text, alphanumeric text, URIs, integers, currency,
+             floats, IP addresses, dates (ISO, long and short formats) and time.
+             Add your own easily.
+           * Inline editing
+           * Support for ROWSPAN and COLSPAN on TH elements.
+           * Support secondary "hidden" sorting (e.g., maintain alphabetical sort when
+             sorting on other criteria).
+           * Extensibility via widget system.
+           * Cross-browser: IE 6.0+, FF 2+, Safari 2.0+, Opera 9.0+, Chrome 5.0+.
+
+    """
+    ).splitlines(keepends=True)
+
+    diagnostics = line_linter(lines)
+    assert len(diagnostics) == 1
+
+    issue = diagnostics[0]
+
+    msg = (
+        "The Multi-Arch field name was used multiple times in this stanza."
+        " Please ensure the field is only used once per stanza. Note that Multi-Arch and"
+        " X[BCS]-Multi-Arch are considered the same field."
+    )
+    assert issue.message == msg
+    assert f"{issue.range}" == "27:0-27:10"
+    assert issue.severity == DiagnosticSeverity.Error
